#ifndef _GAMELIST_H_
#define _GAMELIST_H_

#include <QString>
#include <QTime>
#include <QProcess>
#include <QIcon>
#include <QFile>
#include <QTreeWidget>
#include <QTextStream>

// sort criteria
#define QMC2_SORT_BY_DESCRIPTION	0
#define QMC2_SORT_BY_ROM_STATE		1
#define QMC2_SORT_BY_YEAR		2
#define QMC2_SORT_BY_MANUFACTURER	3

#define QMC2_ROMSTATE_COUNT		5
#define QMC2_ROMSTATE_INT_C		0
#define QMC2_ROMSTATE_INT_M		1
#define QMC2_ROMSTATE_INT_U		2
#define QMC2_ROMSTATE_INT_I		3
#define QMC2_ROMSTATE_INT_N		4
#define QMC2_ROMSTATE_CHAR_C		'0'
#define QMC2_ROMSTATE_CHAR_M		'1'
#define QMC2_ROMSTATE_CHAR_U		'2'
#define QMC2_ROMSTATE_CHAR_I		'3'
#define QMC2_ROMSTATE_CHAR_N		'4'
#define QMC2_ROMSTATE_STRING_C		"0"
#define QMC2_ROMSTATE_STRING_M		"1"
#define QMC2_ROMSTATE_STRING_U		"2"
#define QMC2_ROMSTATE_STRING_I		"3"
#define QMC2_ROMSTATE_STRING_N		"4"

#define QMC2_STATEFILTER_RESPONSIVENESS	1000

class Gamelist : public QObject
{
  Q_OBJECT

  public:
    QIcon qmc2SmallGhostImageIcon;
    QIcon qmc2UnknownImageIcon;
    QIcon qmc2CorrectImageIcon;
    QIcon qmc2MostlyCorrectImageIcon;
    QIcon qmc2IncorrectImageIcon;
    QIcon qmc2NotFoundImageIcon;
    QProcess *loadProc;
    QProcess *verifyProc;
    QString gamelistBuffer;
    QStringList xmlLines;
    static QStringList phraseTranslatorList;
    QTime loadTimer;
    QTime verifyTimer;
    QTime parseTimer;
    QFile romCache;
    QTextStream tsRomCache;
    int numTotalGames;
    int numGames;
    int numCorrectGames;
    int numMostlyCorrectGames;
    int numIncorrectGames;
    int numNotFoundGames;
    int numUnknownGames;
    int numSearchGames;
    bool verifyCurrentOnly;
    QTreeWidgetItem *checkedItem;

    Gamelist(QObject *parent = 0);
    ~Gamelist();

  public slots:
    void load();
    void verify(bool currentOnly = FALSE);
    void save();
    void loadFavorites();
    void saveFavorites();
    void loadPlayHistory();
    void savePlayHistory();
    QString status();

    // process management
    void loadStarted();
    void loadFinished(int, QProcess::ExitStatus);
    void loadReadyReadStandardOutput();
    void loadReadyReadStandardError();
    void loadError(QProcess::ProcessError);
    void loadStateChanged(QProcess::ProcessState);
    void verifyStarted();
    void verifyFinished(int, QProcess::ExitStatus);
    void verifyReadyReadStandardOutput();
    void verifyReadyReadStandardError();
    void verifyError(QProcess::ProcessError);
    void verifyStateChanged(QProcess::ProcessState);

    // internal methods
    QString value(QString, QString, bool translate = FALSE);
    void parse();
    void parseGameDetail(QTreeWidgetItem *);
    void insertAttributeItems(QTreeWidgetItem *, QString, QStringList, QStringList, bool translate = FALSE);
    void enableWidgets(bool enable = TRUE);
    void filter();
    bool loadIcon(QString, QTreeWidgetItem *, bool checkOnly = FALSE, QString *fileName = NULL);
};

class GamelistItem : public QTreeWidgetItem
{
  public:
    GamelistItem(QTreeWidget *parentTreeWidget) : QTreeWidgetItem(parentTreeWidget, QTreeWidgetItem::UserType) {;}
    GamelistItem(QTreeWidgetItem *parentItem) : QTreeWidgetItem(parentItem, QTreeWidgetItem::UserType) {;}

    virtual bool operator<(const QTreeWidgetItem &) const;
};

#endif
