#include <QMessageBox>
#include <QPixmapCache>
#include <QHeaderView>
#include <QTextStream>
#include <QScrollBar>
#include <QProcess>
#include <QTimer>
#include <QTime>
#include <QFile>
#include <QMap>
#include <QStyleFactory>
#include <QBitArray>

#include "macros.h"
#include "qmc2main.h"
#include "options.h"
#include "mameopt.h"
#include "procmgr.h"
#include "gamelist.h"
#include "preview.h"
#include "flyer.h"
#include "dbrowser.h"
#include "about.h"
#include "welcome.h"
#include "imgcheck.h"
#include "sampcheck.h"
#include "unzip.h"

// global variables
MainWindow *qmc2MainWindow = NULL;
Gamelist *qmc2Gamelist = NULL;
Options *qmc2Options = NULL;
QSettings *qmc2Config = NULL;
MameOptions *qmc2GlobalMAMEOptions = NULL;
MameOptions *qmc2MAMEOptions = NULL;
ProcessManager *qmc2ProcessManager = NULL;
Preview *qmc2Preview = NULL;
Flyer *qmc2Flyer = NULL;
About *qmc2About = NULL;
DocBrowser *qmc2DocBrowser = NULL;
Welcome *qmc2Welcome = NULL;
ImageChecker *qmc2ImageChecker = NULL;
SampleChecker *qmc2SampleChecker = NULL;
bool qmc2ReloadActive = FALSE;
bool qmc2ImageCheckActive = FALSE;
bool qmc2SampleCheckActive = FALSE;
bool qmc2EarlyReloadActive = FALSE;
bool qmc2VerifyActive = FALSE;
bool qmc2FilterActive = FALSE;
bool qmc2GuiReady = FALSE;
bool qmc2CleaningUp = FALSE;
bool qmc2StartingUp = TRUE;
bool qmc2EarlyStartup = TRUE;
bool qmc2ScaledPreview = TRUE;
bool qmc2ScaledFlyer = TRUE;
bool qmc2SmoothScaling = FALSE;
bool qmc2UsePreviewFile = FALSE;
bool qmc2UseFlyerFile = FALSE;
bool qmc2UseIconFile = FALSE;
bool qmc2IconsPreloaded = FALSE;
int qmc2GamelistResponsiveness = 100;
QFile *qmc2FrontendLogFile = NULL;
QFile *qmc2MAMELogFile = NULL;
QTextStream qmc2FrontendLogStream;
QTextStream qmc2MAMELogStream;
QTranslator *qmc2Translator = NULL;
QTranslator *qmc2QtTranslator = NULL;
QString qmc2LastFrontendLogMessage = "";
quint64 qmc2FrontendLogMessageRepeatCount = 0;
QString qmc2LastMAMELogMessage = "";
quint64 qmc2MAMELogMessageRepeatCount = 0;
bool qmc2StopParser = FALSE;
QTreeWidgetItem *qmc2CurrentItem = NULL;
QTreeWidgetItem *qmc2LastItem = NULL;
QMenu *qmc2EmulatorMenu = NULL,
      *qmc2GameMenu = NULL,
      *qmc2FavoritesMenu = NULL,
      *qmc2PlayedMenu = NULL,
      *qmc2SearchMenu = NULL;
QMap<QString, QTreeWidgetItem *> qmc2GamelistItemMap;
QMap<QString, QTreeWidgetItem *> qmc2HierarchyItemMap;
QMap<QString, QString> qmc2GamelistNameMap;
QMap<QString, QString> qmc2GamelistDescriptionMap;
QMap<QString, QString> qmc2GamelistStatusMap;
QMap<QString, QStringList> qmc2HierarchyMap;
QMap<QString, QIcon> qmc2IconMap;
int qmc2SortCriteria = QMC2_SORT_BY_DESCRIPTION;
Qt::SortOrder qmc2SortOrder = Qt::AscendingOrder;
QBitArray qmc2Filter;
unzFile qmc2IconFile = NULL;

void MainWindow::log(char logOrigin, QString message)
{
  if ( !qmc2GuiReady )
    return;

  QString timeString = QTime::currentTime().toString("hh:mm:ss.zzz");

  // count subsequent message duplicates
  switch ( logOrigin ) {
    case LOG_QMC2:
      if ( message == qmc2LastFrontendLogMessage ) {
        qmc2FrontendLogMessageRepeatCount++;
        return;
      } else {
        qmc2LastFrontendLogMessage = message;
        if ( qmc2FrontendLogMessageRepeatCount > 0 )
          message.prepend(tr("last message repeated %1 time(s)\n").arg(qmc2FrontendLogMessageRepeatCount) + timeString + ": ");
        qmc2FrontendLogMessageRepeatCount = 0;
      }
      break;

    case LOG_MAME:
      if ( message == qmc2LastMAMELogMessage ) {
        qmc2MAMELogMessageRepeatCount++;
        return;
      } else {
        qmc2LastMAMELogMessage = message;
        if ( qmc2MAMELogMessageRepeatCount > 0 )
          message.prepend(tr("last message repeated %1 time(s)\n").arg(qmc2MAMELogMessageRepeatCount) + timeString + ": ");
        qmc2MAMELogMessageRepeatCount = 0;
      }
      break;

    default:
      break;
  }

  QString msg = timeString + ": " + message;

  switch ( logOrigin ) {
    case LOG_QMC2:
      textBrowserFrontendLog->append(msg);
      if ( !qmc2FrontendLogFile )
        if ( (qmc2FrontendLogFile = new QFile(qmc2Config->value("Frontend/FilesAndDirectories/LogFile", "data/log/qmc2.log").toString(), this)) == NULL )
          return;
      if ( qmc2FrontendLogFile->handle() == -1 )
        if ( qmc2FrontendLogFile->open(QIODevice::WriteOnly | QIODevice::Text) )
          qmc2FrontendLogStream.setDevice(qmc2FrontendLogFile);
        else
          return;
      qmc2FrontendLogStream << msg << "\n";
      qmc2FrontendLogStream.flush();
      break;

    case LOG_MAME:
      textBrowserMAMELog->append(msg);
      if ( !qmc2MAMELogFile )
        if ( (qmc2MAMELogFile = new QFile(qmc2Config->value("MAME/FilesAndDirectories/LogFile", "data/log/mame.log").toString(), this)) == NULL )
          return;
      if ( qmc2MAMELogFile->handle() == -1 )
        if ( qmc2MAMELogFile->open(QIODevice::WriteOnly | QIODevice::Text) )
          qmc2MAMELogStream.setDevice(qmc2MAMELogFile);
        else
          return;
      qmc2MAMELogStream << msg << "\n";
      qmc2MAMELogStream.flush();
      break;

    default:
      break;
  }
}

MainWindow::MainWindow(QWidget *parent)
  : QMainWindow(parent)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::MainWindow(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  setupUi(this);

  qmc2Gamelist = new Gamelist(this);

  statusBar()->setShown(qmc2Config->value("Frontend/GUI/Statusbar").toBool());

  // preview & flyer widgets
  QHBoxLayout *previewLayout = new QHBoxLayout, *flyerLayout = new QHBoxLayout;
  qmc2Preview = new Preview(tabPreview);
  previewLayout->addWidget(qmc2Preview);
  tabPreview->setLayout(previewLayout);
  qmc2Flyer = new Flyer(tabFlyer);
  flyerLayout->addWidget(qmc2Flyer);
  tabFlyer->setLayout(flyerLayout);

  // setup application style
  QString myStyle = qmc2Config->value("Frontend/GUI/Style", tr("Default")).toString();
  if ( myStyle != QObject::tr("Default") )
    setupStyle(myStyle);

  // restore layout
  if ( qmc2Config->value("Frontend/GUI/RestoreLayout").toBool() ) {
    log(LOG_QMC2, tr("restoring main widget layout"));
    move(qmc2Config->value("Frontend/Layout/MainWidget/Position", QPoint(0, 0)).toPoint());
    resize(qmc2Config->value("Frontend/Layout/MainWidget/Size", QSize(800, 600)).toSize());
    QList<int> hSplitterSizes, vSplitterSizes;
    QSize hSplitterSize = qmc2Config->value("Frontend/Layout/MainWidget/hSplitter").toSize();
    QSize vSplitterSize = qmc2Config->value("Frontend/Layout/MainWidget/vSplitter").toSize();
    if ( hSplitterSize.width() > 0 || hSplitterSize.height() > 0 )
      hSplitterSizes << hSplitterSize.width() << hSplitterSize.height();
    else
      hSplitterSizes << 100 << 100;
    if ( vSplitterSize.width() > 0 || vSplitterSize.height() > 0 )
      vSplitterSizes << vSplitterSize.width() << vSplitterSize.height();
    else
      vSplitterSizes << 100 << 100;
    hSplitter->setSizes(hSplitterSizes);
    vSplitter->setSizes(vSplitterSizes);
    tabWidgetGamelist->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/GamelistTab", 0).toInt());
    tabWidgetGameDetail->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/GameDetailTab", 0).toInt());
    tabWidgetLogsAndEmulators->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/LogsAndEmulatorsTab", 0).toInt());
    int i;
    QVariantList defaultGamelistColumnWidths,
                 gamelistColumnWidths;
    defaultGamelistColumnWidths << 300 << 50 << 50 << 100;
    gamelistColumnWidths = qmc2Config->value("Frontend/Layout/MainWidget/GamelistColumnWidths", defaultGamelistColumnWidths).toList();
    for (i = 0; i < gamelistColumnWidths.count(); i++)
      treeWidgetGamelist->header()->resizeSection(i, gamelistColumnWidths[i].toInt());
    treeWidgetGamelist->header()->setDefaultAlignment(Qt::AlignLeft);
    QVariantList defaultHierarchyColumnWidths,
                 hierarchyColumnWidths;
    defaultHierarchyColumnWidths << 300 << 50 << 50 << 100;
    hierarchyColumnWidths = qmc2Config->value("Frontend/Layout/MainWidget/HierarchyColumnWidths", defaultHierarchyColumnWidths).toList();
    for (i = 0; i < hierarchyColumnWidths.count(); i++)
      treeWidgetHierarchy->header()->resizeSection(i, hierarchyColumnWidths[i].toInt());
    treeWidgetHierarchy->header()->setDefaultAlignment(Qt::AlignLeft);
    if ( qmc2Config->value("Frontend/Layout/ImageChecker/Visible").toBool() ) {
      on_actionCheckPreviews_activated();
      qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(qmc2Config->value("Frontend/ImageChecker/CurrentTab", 0).toInt());
    }
    if ( qmc2Config->value("Frontend/Layout/SampleChecker/Visible").toBool() )
      on_actionCheckSamples_activated();
  }

  // context menus
  QAction *action;
  QString s;

  qmc2EmulatorMenu = new QMenu(0);
  s = tr("Terminate selected emulator(s) (sends TERM signal to emulator process(es))");
  action = qmc2EmulatorMenu->addAction(tr("&Terminate"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/terminate.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_terminateEmulator_triggered()));
  s = tr("Kill selected emulator(s) (sends KILL signal to emulator process(es))");
  action = qmc2EmulatorMenu->addAction(tr("&Kill"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_killEmulator_triggered()));

  qmc2GameMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2GameMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2GameMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));
  qmc2GameMenu->addSeparator();
  s = tr("Check current game's ROM state");
  action = qmc2GameMenu->addAction(tr("Check &ROM state"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/rom.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionCheckCurrentROM_activated()));

  qmc2SearchMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2SearchMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2SearchMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));

  qmc2FavoritesMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2FavoritesMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  qmc2FavoritesMenu->addSeparator();
  s = tr("Remove from favorites");
  action = qmc2FavoritesMenu->addAction(tr("&Remove"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/editdelete.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_removeFromFavorites_triggered()));
  s = tr("Clear all favorites");
  action = qmc2FavoritesMenu->addAction(tr("&Clear"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_clearAllFavorites_triggered()));
  s = tr("Save favorites now");
  action = qmc2FavoritesMenu->addAction(tr("&Save"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/filesave.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_saveFavorites_triggered()));

  qmc2PlayedMenu = new QMenu(0);
  action = qmc2PlayedMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2PlayedMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));
  qmc2PlayedMenu->addSeparator();
  s = tr("Remove from played");
  action = qmc2PlayedMenu->addAction(tr("&Remove"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/editdelete.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_removeFromPlayed_triggered()));
  s = tr("Clear all played");
  action = qmc2PlayedMenu->addAction(tr("&Clear"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_clearAllPlayed_triggered()));
  s = tr("Save play-history now");
  action = qmc2PlayedMenu->addAction(tr("&Save"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/filesave.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_savePlayed_triggered()));

  // other actions
  connect(actionViewFullDetail, SIGNAL(triggered()), this, SLOT(viewFullDetail()));
  connect(actionViewParentClones, SIGNAL(triggered()), this, SLOT(viewParentClones()));
  connect(comboBoxViewSelect, SIGNAL(currentIndexChanged(int)), stackedWidgetView, SLOT(setCurrentIndex(int)));

  treeWidgetGamelist->setFocus();
  comboBoxViewSelect->setCurrentIndex(qmc2Config->value("Frontend/GUI/GamelistView", QMC2_GAMELIST_INDEX).toInt());

  QTimer::singleShot(0, this, SLOT(init()));
}

MainWindow::~MainWindow()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::~MainWindow()");
#endif
}

void MainWindow::on_actionPlay_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionPlay_activated()");
#endif

  if ( qmc2EarlyReloadActive )
    return;

  if ( !qmc2CurrentItem )
    return;

  qmc2LastItem = NULL;
  on_tabWidgetGameDetail_currentChanged(QMC2_MAMECONFIG_INDEX);

  QStringList args;
  QString sectionTitle;
  foreach ( sectionTitle, qmc2MAMEOptions->optionsMap.keys() ) {
    int i;
    for (i = 0; i < qmc2MAMEOptions->optionsMap[sectionTitle].count(); i++) {
      MameOption option = qmc2MAMEOptions->optionsMap[sectionTitle][i];
      QString globalOptionKey = "MAME/Configuration/Global/" + option.name;
      switch ( option.type ) {
        case MAMEOPT_TYPE_INT: {
          int  v = option.value.toInt();
          int dv = option.dvalue.toInt();
          int gv = qmc2Config->value(globalOptionKey, dv).toInt();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << QString("%1").arg(v);
          break;
        }

        case MAMEOPT_TYPE_FLOAT: {
          double  v = option.value.toDouble();
          double dv = option.dvalue.toDouble();
          double gv = qmc2Config->value(globalOptionKey, dv).toDouble();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << QString("%1").arg(v);
          break;
        }

        case MAMEOPT_TYPE_BOOL: {
          QString  v = option.value;
          QString dv = option.dvalue;
          QString gv = qmc2Config->value(globalOptionKey, dv).toString();
          if ( !option.valid )
            v = gv;
          if ( v != dv ) {
            if ( v == "true" )
              args << QString("-%1").arg(option.name);
            else
              args << QString("-no%1").arg(option.name);
          }
          break;
        }

        case MAMEOPT_TYPE_STRING:
        default: {
          QString  v = option.value;
          QString dv = option.dvalue;
          QString gv = qmc2Config->value(globalOptionKey, dv).toString();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << v.replace("~", "$HOME");
          break;
        }
      }
    }
  }

  // start game
  QString gameName = qmc2CurrentItem->child(0)->text(1);
  args << gameName;
  QString command = qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString();
  qmc2ProcessManager->process(qmc2ProcessManager->start(command, args));

  // add game to played list
  QList<QListWidgetItem *> matches = listWidgetPlayed->findItems(qmc2GamelistItemMap[gameName]->text(0), Qt::MatchExactly);
  QListWidgetItem *playedItem;
  if ( matches.count() > 0 ) {
    playedItem = listWidgetPlayed->takeItem(listWidgetPlayed->row(matches[0]));
  } else {
    playedItem = new QListWidgetItem();
    playedItem->setText(qmc2GamelistItemMap[gameName]->text(0));
  }
  listWidgetPlayed->insertItem(0, playedItem);
  listWidgetPlayed->setCurrentItem(playedItem);
}

void MainWindow::on_actionToFavorites_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionToFavorites_activated()");
#endif

  if ( !qmc2CurrentItem )
    return;

  QString itemText = qmc2CurrentItem->text(0);

  if ( itemText == tr("Waiting for data...") )
    return;

  QList<QListWidgetItem *> matches = listWidgetFavorites->findItems(itemText, Qt::MatchExactly);
  if ( matches.count() <= 0 ) {
    QListWidgetItem *item = new QListWidgetItem(listWidgetFavorites);
    item->setText(itemText);
    listWidgetFavorites->sortItems();
  }
}

void MainWindow::on_actionReload_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionReload_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
    return;
  }

  if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("please wait for ROM verification to finish and try again"));
    return;
  }

  if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
    return;
  }

  if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
    return;
  }

  if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("gamelist reload already active"));
  } else {
    if ( qmc2Config->value("Frontend/GUI/SaveGameSelection").toBool() && !qmc2StartingUp ) {
      if ( qmc2CurrentItem ) {
        log(LOG_QMC2, tr("saving game selection"));
        qmc2Config->setValue("Frontend/Gamelist/SelectedGame", qmc2CurrentItem->text(0));
      } else
        qmc2Config->remove("Frontend/Gamelist/SelectedGame");
    }
    qmc2Gamelist->load();
  }
}

void MainWindow::on_actionExitStop_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionExitStop_activated()");
#endif

  close();
}

void MainWindow::on_actionCheckCurrentROM_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckCurrentROM_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
  } else if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("ROM verification already active"));
  } else if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
  } else if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
  } else if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
  } else
    qmc2Gamelist->verify(TRUE);
}

void MainWindow::on_actionCheckROMs_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckROMs_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
  } else if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("ROM verification already active"));
  } else if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
  } else if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
  } else if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
  } else {
    switch ( QMessageBox::question(this,
                                   tr("Confirm"),
                                   tr("The ROM verification process may be very time-consuming.\nIt will overwrite existing cached data.\n\nDo you really want to check all ROM states now?"),
                                   QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
        case QMessageBox::Yes:
          qmc2Gamelist->verify();
          break;

        default:
          break;
    }
  }
}

void MainWindow::on_actionCheckSamples_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckSamples_activated()");
#endif

  if ( !qmc2SampleChecker )
    qmc2SampleChecker = new SampleChecker(0);

  if ( qmc2SampleChecker->isHidden() )
    qmc2SampleChecker->show();
  else if ( qmc2SampleChecker->isMinimized() )
    qmc2SampleChecker->showNormal();
  qmc2SampleChecker->raise();
}

void MainWindow::on_actionCheckPreviews_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckPreviews_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_PREVIEW_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();
  qmc2ImageChecker->raise();
}

void MainWindow::on_actionCheckFlyers_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckFlyers_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_FLYER_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();
  qmc2ImageChecker->raise();
}

void MainWindow::on_actionCheckIcons_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckIcons_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_ICON_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();
  qmc2ImageChecker->raise();
}

void MainWindow::on_actionROMAdministration_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionROMAdministration_activated()");
#endif

  log(LOG_QMC2, tr("sorry, this feature is not yet available"));
}

void MainWindow::on_actionClearImageCache_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionClearImageCache_activated()");
#endif

  QPixmapCache::clear();
  log(LOG_QMC2, tr("image cache cleared"));
}

void MainWindow::on_actionClearIconCache_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionClearIconCache_activated()");
#endif

  qmc2IconMap.clear();
  qmc2IconsPreloaded = FALSE;
  log(LOG_QMC2, tr("icon cache cleared"));
}

void MainWindow::on_actionRecreateTemplateMap_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionRecreateTemplateMap_activated()");
#endif

  qmc2GlobalMAMEOptions->createTemplateMap();
}

void MainWindow::on_actionOptions_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionOptions_activated()");
#endif

  if ( qmc2Options->isHidden() )
    qmc2Options->show();
  else if ( qmc2Options->isMinimized() )
    qmc2Options->showNormal();
  qmc2Options->raise();
}

void MainWindow::on_actionDocumentation_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionDocumentation_activated()");
#endif

  if ( !qmc2DocBrowser ) {
    qmc2DocBrowser = new DocBrowser(this);
    QStringList searchPaths;
    searchPaths << qmc2Config->value("Frontend/FilesAndDirectories/DataDirectory").toString() +
                   "doc/html/" +
                   qmc2Config->value("Frontend/GUI/Language").toString();
    qmc2DocBrowser->manualBrowser->setSearchPaths(searchPaths);
    qmc2DocBrowser->manualBrowser->setSource(QUrl("index.html"));
  }

  qmc2DocBrowser->show();
  qmc2DocBrowser->raise();
}

void MainWindow::on_actionAbout_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionAbout_activated()");
#endif

  if ( !qmc2About )
    qmc2About = new About(this);

  qmc2About->show();
  qmc2About->raise();
}

void MainWindow::on_actionAboutQt_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionAboutQt_activated()");
#endif

  QMessageBox::aboutQt(this, tr("About Qt"));
}

void MainWindow::on_comboBoxSearch_textChanged(QString pattern)
{
  // easy pattern match
  pattern.replace(' ', "* *");
  pattern = "*" + pattern + "*";
  pattern.replace(QString("*^"), "");
  pattern.replace(QString("$*"), "");

  listWidgetSearch->clear();
  QList<QTreeWidgetItem *> matches = treeWidgetGamelist->findItems(pattern, Qt::MatchContains | Qt::MatchWildcard);
  
  int i;
  for (i = 0; i < matches.count(); i++) {
    QListWidgetItem *item = new QListWidgetItem(listWidgetSearch);
    item->setText(matches.at(i)->text(0));
  }

  qmc2Gamelist->numSearchGames = matches.count();
  labelGamelistStatus->setText(qmc2Gamelist->status());

#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_comboBoxSearch_textChanged(QString pattern = \"" + pattern + "\"): matches.count() = " + QString::number(matches.count()));
#endif

  QTimer::singleShot(0, this, SLOT(checkCurrentSearchSelection()));
}

void MainWindow::on_comboBoxSearch_activated(QString pattern)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_comboBoxSearch_activated(QString pattern = \"" + pattern + "\")");
#endif

  on_comboBoxSearch_textChanged(pattern);
  if ( tabWidgetGamelist->currentWidget() != tabSearch )
    tabWidgetGamelist->setCurrentWidget(tabSearch);
}

void MainWindow::on_listWidgetSearch_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_currentTextChanged(QString s = \""+ s + "\")");
#endif

  QList<QTreeWidgetItem *> matches = treeWidgetGamelist->findItems(s, Qt::MatchExactly);
  if ( matches.count() > 0 ) {
    QTreeWidgetItem *matchItem = matches.at(0);
    treeWidgetGamelist->setCurrentItem(matchItem);
  }
#ifdef QMC2_DEBUG
  else
    log(LOG_QMC2, "DEBUG: ERROR: no match found (?)");
#endif
}

void MainWindow::on_listWidgetSearch_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  if ( item == NULL )
    return;

  QList<QTreeWidgetItem *> matches = treeWidgetGamelist->findItems(item->text(), Qt::MatchExactly);
  if ( matches.count() > 0 ) {
    QTreeWidgetItem *matchItem = matches.at(0);
    tabWidgetGamelist->setCurrentIndex(0);
    treeWidgetGamelist->setCurrentItem(matchItem);
    if ( !qmc2ReloadActive )
      treeWidgetGamelist->expandItem(matchItem);
  }
  else
    log(LOG_QMC2, tr("ERROR: no match found"));
}

void MainWindow::on_listWidgetFavorites_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_currentTextChanged(QString s = \""+ s + "\")");
#endif

  on_listWidgetSearch_currentTextChanged(s);
}

void MainWindow::on_listWidgetFavorites_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_listWidgetSearch_itemActivated(item);
}

void MainWindow::on_listWidgetPlayed_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_currentTextChanged(QString s = \""+ s + "\")");
#endif

  on_listWidgetSearch_currentTextChanged(s);
}

void MainWindow::on_listWidgetPlayed_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_listWidgetSearch_itemActivated(item);
}

void MainWindow::on_tabWidgetGamelist_currentChanged(int currentIndex)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_tabWidgetGamelist_currentChanged(int i = " + QString::number(currentIndex) + ")");
#endif

  switch ( currentIndex ) {
    case 0:
      QTimer::singleShot(0, this, SLOT(scrollToCurrentItem()));
      break;

    case 1:
      QTimer::singleShot(0, this, SLOT(checkCurrentSearchSelection()));
      break;

    case 2:
      QTimer::singleShot(0, this, SLOT(checkCurrentFavoritesSelection()));
      break;

    case 3:
      QTimer::singleShot(0, this, SLOT(checkCurrentPlayedSelection()));
      break;

    default:
      break;
  }
}

void MainWindow::scrollToCurrentItem()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::scrollToCurrentItem()");
#endif

  QTreeWidgetItem *ci = treeWidgetGamelist->currentItem();

  if ( qmc2CurrentItem )
    ci = qmc2CurrentItem;

  if ( ci ) {
    if ( ci->text(0) == tr("Waiting for data...") )
      return;

    switch ( stackedWidgetView->currentIndex() ) {
      case QMC2_HIERARCHY_INDEX:
        ci = qmc2HierarchyItemMap[ci->child(0)->text(1)];
        if ( ci ) {
          treeWidgetHierarchy->clearSelection();
          treeWidgetHierarchy->setCurrentItem(ci);
          treeWidgetHierarchy->scrollToItem(ci, QAbstractItemView::PositionAtTop);
        }
        break;

      case QMC2_GAMELIST_INDEX:
      default:
        treeWidgetGamelist->clearSelection();
        treeWidgetGamelist->setCurrentItem(ci);
        treeWidgetGamelist->scrollToItem(ci, QAbstractItemView::PositionAtTop);
        break;
    }
    ci->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentSearchSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentSearchSelection()");
#endif

  if ( !qmc2CurrentItem )
    return;

  listWidgetSearch->clearFocus();
  listWidgetSearch->clearSelection();
  QList<QListWidgetItem *> searchMatches = listWidgetSearch->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( searchMatches.count() > 0 ) {
    listWidgetSearch->setCurrentItem(searchMatches[0]);
    listWidgetSearch->scrollToItem(searchMatches[0], QAbstractItemView::PositionAtTop);
    searchMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentFavoritesSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentFavoritesSelection()");
#endif

  if ( !qmc2CurrentItem )
    return;

  listWidgetFavorites->clearFocus();
  listWidgetFavorites->clearSelection();
  QList<QListWidgetItem *> favoritesMatches = listWidgetFavorites->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( favoritesMatches.count() > 0 ) {
    listWidgetFavorites->setCurrentItem(favoritesMatches[0]);
    listWidgetFavorites->scrollToItem(favoritesMatches[0], QAbstractItemView::PositionAtTop);
    favoritesMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentPlayedSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentPlayedSelection()");
#endif

  if ( !qmc2CurrentItem )
    return;

  listWidgetPlayed->clearFocus();
  listWidgetPlayed->clearSelection();
  QList<QListWidgetItem *> playedMatches = listWidgetPlayed->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( playedMatches.count() > 0 ) {
    listWidgetPlayed->setCurrentItem(playedMatches[0]);
    listWidgetPlayed->scrollToItem(playedMatches[0], QAbstractItemView::PositionAtTop);
    playedMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::on_tabWidgetGameDetail_currentChanged(int currentIndex)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_tabWidgetGameDetail_currentChanged(int currentIndex = " + QString::number(currentIndex) + ")");
#endif

  if ( !qmc2CurrentItem )
    return;
 
  if ( qmc2EarlyReloadActive )
    return;

  if ( qmc2CurrentItem != qmc2LastItem ) {
    tabMAME->setUpdatesEnabled(FALSE);
    if ( qmc2MAMEOptions ) {
      qmc2MAMEOptions->save();
      QLayout *hbl = tabMAME->layout();
      delete hbl;
      delete qmc2MAMEOptions;
      qmc2MAMEOptions = NULL;
    }

    if ( currentIndex == QMC2_MAMECONFIG_INDEX ) {
      QString gameName = qmc2CurrentItem->child(0)->text(1);
      log(LOG_QMC2, tr("processing MAME configuration for %1").arg(qmc2CurrentItem->text(0)));
      QHBoxLayout *layout = new QHBoxLayout;
      qmc2MAMEOptions = new MameOptions("MAME/Configuration/" + gameName, tabMAME);
      qmc2MAMEOptions->load();
      layout->addWidget(qmc2MAMEOptions);
      tabMAME->setLayout(layout);
      qmc2MAMEOptions->show();
      qmc2LastItem = qmc2CurrentItem;
    }
    tabMAME->setUpdatesEnabled(TRUE);
  }
}

void MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_actionPlay_activated();
}

void MainWindow::on_treeWidgetHierarchy_itemActivated(QTreeWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_actionPlay_activated();
}

void MainWindow::on_treeWidgetGamelist_itemExpanded(QTreeWidgetItem *item)
{
  if ( qmc2ReloadActive ) {
    treeWidgetGamelist->collapseItem(item);
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
    return;
  }

#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemExpanded(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  if ( item->child(0) ) {
    if ( item->child(0)->text(0) == tr("Waiting for data...") )
      qmc2Gamelist->parseGameDetail(item);
  }
}

void MainWindow::on_treeWidgetGamelist_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_currentItemChanged(QTreeWidgetItem *current = 0x" + QString::number((ulong)current, 16) + ", QTreeWidgetItem *previous = 0x" + QString::number((ulong)previous, 16) + ")");
#endif

  on_tabWidgetGameDetail_currentChanged(tabWidgetGameDetail->currentIndex());
}

void MainWindow::on_treeWidgetHierarchy_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_currentItemChanged(QTreeWidgetItem *current = 0x" + QString::number((ulong)current, 16) + ", QTreeWidgetItem *previous = 0x" + QString::number((ulong)previous, 16) + ")");
#endif

  on_tabWidgetGameDetail_currentChanged(tabWidgetGameDetail->currentIndex());
}

void MainWindow::on_treeWidgetGamelist_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemSelectionChanged()");
#endif

  QList<QTreeWidgetItem *>selected = treeWidgetGamelist->selectedItems();
  if ( selected.count() > 0 ) {
    QTreeWidgetItem *topLevelItem = selected.at(0);
    while ( topLevelItem->parent() )
      topLevelItem = topLevelItem->parent();
    if ( topLevelItem != qmc2CurrentItem ) {
      qmc2CurrentItem = topLevelItem;
      qmc2Preview->repaint();
      qmc2Flyer->repaint();
    }
  } else
    qmc2CurrentItem = NULL;
}

void MainWindow::on_treeWidgetHierarchy_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_itemSelectionChanged()");
#endif

  QList<QTreeWidgetItem *>selected = treeWidgetHierarchy->selectedItems();
  if ( selected.count() > 0 ) {
    if ( selected.at(0)->text(0) == tr("Waiting for data...") )
      return;
    treeWidgetGamelist->clearSelection();
    qmc2GamelistItemMap[qmc2GamelistNameMap[selected.at(0)->text(0)]]->setSelected(TRUE);
  }
}

void MainWindow::on_textBrowserFrontendLog_textChanged()
{
  textBrowserFrontendLog->verticalScrollBar()->setSliderPosition(textBrowserFrontendLog->verticalScrollBar()->maximum());
  textBrowserFrontendLog->horizontalScrollBar()->setSliderPosition(textBrowserFrontendLog->horizontalScrollBar()->minimum());
}

void MainWindow::on_textBrowserMAMELog_textChanged()
{
  textBrowserMAMELog->verticalScrollBar()->setSliderPosition(textBrowserMAMELog->verticalScrollBar()->maximum());
  textBrowserMAMELog->horizontalScrollBar()->setSliderPosition(textBrowserMAMELog->horizontalScrollBar()->minimum());
}

void MainWindow::on_listWidgetEmulators_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetEmulators_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetEmulators->itemAt(p);
  if ( item ) {
    listWidgetEmulators->setItemSelected(item, TRUE);
    qmc2EmulatorMenu->move(listWidgetEmulators->viewport()->mapToGlobal(p));
    qmc2EmulatorMenu->show();
  }
}

void MainWindow::action_terminateEmulator_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_terminateEmulator_triggered()");
#endif

  QList<QListWidgetItem *> sl = listWidgetEmulators->selectedItems();
  int i;
  for (i = 0; i < sl.count(); i++) {
    QString s = sl.at(i)->text().remove("#");
    s.truncate(s.indexOf(":"));
    qmc2ProcessManager->terminate(s.toInt());
  }

  if ( sl.count() == 0 )
    if ( listWidgetEmulators->currentItem() ) {
      QString s = listWidgetEmulators->currentItem()->text().remove("#");
      s.truncate(s.indexOf(":"));
      qmc2ProcessManager->terminate(s.toInt());
    }
}

void MainWindow::action_killEmulator_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_killEmulator_triggered()");
#endif

  QList<QListWidgetItem *> sl = listWidgetEmulators->selectedItems();
  int i;
  for (i = 0; i < sl.count(); i++) {
    QString s = sl.at(i)->text().remove("#");
    s.truncate(s.indexOf(":"));
    qmc2ProcessManager->kill(s.toInt());
  }

  if ( sl.count() == 0 )
    if ( listWidgetEmulators->currentItem() ) {
      QString s = listWidgetEmulators->currentItem()->text().remove("#");
      s.truncate(s.indexOf(":"));
      qmc2ProcessManager->kill(s.toInt());
    }
}

void MainWindow::action_removeFromFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_removeFromFavorites_triggered()");
#endif

  QListWidgetItem *i = listWidgetFavorites->currentItem();
  if ( i ) {
    QListWidgetItem *item = listWidgetFavorites->takeItem(listWidgetFavorites->row(i));
    delete item;
  }
}

void MainWindow::action_clearAllFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_clearAllFavorites_triggered()");
#endif

  listWidgetFavorites->clear();
}

void MainWindow::action_saveFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_saveFavorites_triggered()");
#endif

  qmc2Gamelist->saveFavorites();
}

void MainWindow::action_removeFromPlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_removeFromPlayed_triggered()");
#endif

  QListWidgetItem *i = listWidgetPlayed->currentItem();
  if ( i ) {
    QListWidgetItem *item = listWidgetPlayed->takeItem(listWidgetPlayed->row(i));
    delete item;
  }
}

void MainWindow::action_clearAllPlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_clearAllPlayed_triggered()");
#endif

  listWidgetPlayed->clear();
}

void MainWindow::action_savePlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_savePlayed_triggered()");
#endif

  qmc2Gamelist->savePlayHistory();
}

void MainWindow::on_listWidgetSearch_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetSearch->itemAt(p);
  if ( item ) {
    listWidgetSearch->setItemSelected(item, TRUE);
    qmc2SearchMenu->move(listWidgetSearch->viewport()->mapToGlobal(p));
    qmc2SearchMenu->show();
  }
}

void MainWindow::on_listWidgetFavorites_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetFavorites->itemAt(p);
  if ( item ) {
    listWidgetFavorites->setItemSelected(item, TRUE);
    qmc2FavoritesMenu->move(listWidgetFavorites->viewport()->mapToGlobal(p));
    qmc2FavoritesMenu->show();
  }
}

void MainWindow::on_listWidgetPlayed_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetPlayed->itemAt(p);
  if ( item ) {
    listWidgetPlayed->setItemSelected(item, TRUE);
    qmc2PlayedMenu->move(listWidgetPlayed->viewport()->mapToGlobal(p));
    qmc2PlayedMenu->show();
  }
}

void MainWindow::on_treeWidgetGamelist_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QTreeWidgetItem *item = treeWidgetGamelist->itemAt(p);
  if ( !item )
    return;
  if ( item->text(0) == tr("Waiting for data...") )
    return;
  if ( item ) {
    treeWidgetGamelist->setItemSelected(item, TRUE);
    qmc2GameMenu->move(treeWidgetGamelist->viewport()->mapToGlobal(p));
    qmc2GameMenu->show();
  }
}

void MainWindow::on_treeWidgetHierarchy_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QTreeWidgetItem *item = treeWidgetHierarchy->itemAt(p);
  if ( !item )
    return;
  if ( item->text(0) == tr("Waiting for data...") )
    return;
  if ( item ) {
    treeWidgetHierarchy->setItemSelected(item, TRUE);
    qmc2GameMenu->move(treeWidgetHierarchy->viewport()->mapToGlobal(p));
    qmc2GameMenu->show();
  }
}

void MainWindow::on_stackedWidgetView_currentChanged(int index)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_tackedWidgetView_currentChanged(int index = %1)").arg(index));
#endif

  if ( !qmc2CurrentItem )
    return;

  if ( qmc2CurrentItem->text(0) == tr("Waiting for data...") )
    return;

  QTreeWidgetItem *hierarchyItem;
  switch ( index ) {
    case QMC2_HIERARCHY_INDEX:
      hierarchyItem = qmc2HierarchyItemMap[qmc2CurrentItem->child(0)->text(1)];
      treeWidgetHierarchy->clearSelection();
      if ( hierarchyItem ) {
        treeWidgetHierarchy->setCurrentItem(hierarchyItem);
        treeWidgetHierarchy->scrollToItem(hierarchyItem, QAbstractItemView::PositionAtTop);
        hierarchyItem->setSelected(TRUE);
      }
      break;

    case QMC2_GAMELIST_INDEX:
    default:
      scrollToCurrentItem();
      break;
  }
}

void MainWindow::closeEvent(QCloseEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::closeEvent(QCloseEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( qmc2CleaningUp ) {
    e->ignore();
    return;
  }

  if ( qmc2ReloadActive || qmc2VerifyActive || qmc2FilterActive || qmc2ImageCheckActive || qmc2SampleCheckActive ) {
    qmc2StopParser = TRUE;
    log(LOG_QMC2, tr("stopping current processing upon user request"));
    e->ignore();
    return;
  }

  if ( !qmc2Options->applied ) {
    switch ( QMessageBox::question(this, tr("Confirm"), tr("Your configuration changes have not been applied yet.\nReally quit?"), QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
      case QMessageBox::No:
        e->ignore();
        return;
        break;

      default:
        break;
    }
  }
 
  bool doKillEmulators = qmc2Config->value("Frontend/GUI/KillEmulatorsOnExit").toBool();
  if ( qmc2ProcessManager->procMap.count() > 0 && !doKillEmulators ) {
    switch ( QMessageBox::question(this, tr("Confirm"), tr("There are one or more emulators still running.\nShould they be killed on exit?"), QMessageBox::Yes, QMessageBox::No, QMessageBox::Cancel) ) {
      case QMessageBox::Yes:
        doKillEmulators = TRUE;
        break;

      case QMessageBox::No:
        doKillEmulators = FALSE;
        break;

      case QMessageBox::Cancel:
        e->ignore();
        return;
        break;

      default:
        break;
    }
  }

  qmc2CleaningUp = TRUE;
  log(LOG_QMC2, tr("cleaning up"));

  qmc2Gamelist->saveFavorites();
  qmc2Gamelist->savePlayHistory();

  qmc2Config->setValue("Frontend/GUI/GamelistView", comboBoxViewSelect->currentIndex());
  if ( qmc2Config->value("Frontend/GUI/SaveGameSelection").toBool() ) {
    QList<QTreeWidgetItem *> sl = treeWidgetGamelist->selectedItems();
    if ( sl.count() > 0 ) {
      log(LOG_QMC2, tr("saving game selection"));
      qmc2Config->setValue("Frontend/Gamelist/SelectedGame", sl.at(0)->text(0));
    } else
      qmc2Config->remove("Frontend/Gamelist/SelectedGame");
  }

  if ( qmc2Config->value("Frontend/GUI/SaveLayout").toBool() ) {
    log(LOG_QMC2, tr("saving main widget layout"));
    qmc2Config->setValue("Frontend/Layout/MainWidget/Position", pos());
    qmc2Config->setValue("Frontend/Layout/MainWidget/Size", size());
    qmc2Config->setValue("Frontend/Layout/MainWidget/hSplitter", QSize(hSplitter->sizes().at(0), hSplitter->sizes().at(1)));
    qmc2Config->setValue("Frontend/Layout/MainWidget/vSplitter", QSize(vSplitter->sizes().at(0), vSplitter->sizes().at(1)));
    qmc2Config->setValue("Frontend/Layout/MainWidget/GamelistTab", tabWidgetGamelist->currentIndex());
    qmc2Config->setValue("Frontend/Layout/MainWidget/GameDetailTab", tabWidgetGameDetail->currentIndex());
    qmc2Config->setValue("Frontend/Layout/MainWidget/LogsAndEmulatorsTab", tabWidgetLogsAndEmulators->currentIndex());
    int i;
    QVariantList gamelistColumnWidths;
    for (i = 0; i < treeWidgetGamelist->header()->count(); i++)
      gamelistColumnWidths << treeWidgetGamelist->header()->sectionSize(i);
    qmc2Config->setValue("Frontend/Layout/MainWidget/GamelistColumnWidths", gamelistColumnWidths);
    QVariantList hierarchyColumnWidths;
    for (i = 0; i < treeWidgetHierarchy->header()->count(); i++)
      hierarchyColumnWidths << treeWidgetHierarchy->header()->sectionSize(i);
    qmc2Config->setValue("Frontend/Layout/MainWidget/HierarchyColumnWidths", hierarchyColumnWidths);
  }

  if ( qmc2MAMEOptions ) {
    log(LOG_QMC2, tr("destroying current MAME options"));
    qmc2MAMEOptions->save();
    delete qmc2MAMEOptions;
  }
  log(LOG_QMC2, tr("destroying global MAME options"));
  delete qmc2GlobalMAMEOptions;
  log(LOG_QMC2, tr("destroying gamelist"));
  delete qmc2Gamelist;
  log(LOG_QMC2, tr("destroying preview"));
  delete qmc2Preview;
  log(LOG_QMC2, tr("destroying flyer"));
  delete qmc2Flyer;
  if ( qmc2About ) {
    log(LOG_QMC2, tr("destroying about dialog"));
    delete qmc2About;
  }
  if ( qmc2DocBrowser ) {
    log(LOG_QMC2, tr("destroying documentation browser"));
    delete qmc2DocBrowser;
  }
  if ( qmc2ImageChecker ) {
    log(LOG_QMC2, tr("destroying image checker"));
    qmc2ImageChecker->close();
    delete qmc2ImageChecker;
  }
  if ( qmc2SampleChecker ) {
    log(LOG_QMC2, tr("destroying sample checker"));
    qmc2SampleChecker->close();
    delete qmc2SampleChecker;
  }

  log(LOG_QMC2, tr("destroying process manager"));
  if ( qmc2ProcessManager->procMap.count() > 0 ) {
    if ( doKillEmulators ) {
      log(LOG_QMC2, tr("killing %1 running emulator(s) on exit").arg(qmc2ProcessManager->procMap.count()));
      delete qmc2ProcessManager;
    } else
      log(LOG_QMC2, tr("keeping %1 running emulator(s) alive").arg(qmc2ProcessManager->procMap.count()));
  } else 
    delete qmc2ProcessManager;

  log(LOG_QMC2, tr("so long and thanks for all the fish"));

  delete qmc2Options;
  e->accept();
}

void MainWindow::showEvent(QShowEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::showEvent(QShowEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  e->accept();
}

void MainWindow::hideEvent(QHideEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::hideEvent(QHideEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  e->accept();
}

void MainWindow::init()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::init()");
#endif

  treeWidgetGamelist->header()->setClickable(FALSE);
  treeWidgetHierarchy->header()->setClickable(FALSE);
  on_actionReload_activated();
  qmc2EarlyStartup = FALSE;
  qmc2GhostImagePixmap.load(":/data/img/ghost.png");
}

void MainWindow::setupStyle(QString styleName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::setupStyle(QString &styleName = %1").arg(styleName));
#endif
  QApplication::setStyle(QStyleFactory::create(styleName));
  QApplication::setPalette(QApplication::style()->standardPalette());
}

void MainWindow::viewFullDetail()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::viewFullDetail()");
#endif

  comboBoxViewSelect->setCurrentIndex(QMC2_GAMELIST_INDEX);
}

void MainWindow::viewParentClones()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::viewParentClones()");
#endif

  comboBoxViewSelect->setCurrentIndex(QMC2_HIERARCHY_INDEX);
}

void myQtMessageHandler(QtMsgType type, const char *msg)
{
  if ( !qmc2GuiReady )
    return;

  switch ( type ) {
    case QtDebugMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtDebugMsg: " + QString(msg));
      break;

    case QtWarningMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtWarningMsg: " + QString(msg));
      break;

    case QtCriticalMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtCriticalMsg: " + QString(msg));
      break;

    case QtFatalMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtFatalMsg: " + QString(msg));
      break;

    default:
      break;
  }
}

int main(int argc, char *argv[])
{
  qInstallMsgHandler(myQtMessageHandler);
  QApplication app(argc, argv);
  Q_INIT_RESOURCE(qmc2);
  app.setWindowIcon(QIcon(QString::fromUtf8(":/data/img/mame.png")));

  qmc2Welcome = new Welcome(0);
  int checkReturn = qmc2Welcome->exec();
  qmc2Welcome->close();
  delete qmc2Welcome;
  if ( checkReturn != QDialog::Accepted )
    exit(1);

  qmc2Options = new Options(0);
  qmc2Config = qmc2Options->config;
  qmc2ProcessManager = new ProcessManager(0);
  qmc2MainWindow = new MainWindow(0);
  qmc2GuiReady = TRUE;

  qmc2Options->apply();

  qmc2MainWindow->show();
  qmc2MainWindow->raise();
  qmc2Options->raise();

  // greeting
  QString greeting = QObject::tr("M.A.M.E. Catalog / Launcher II v") + QString::number(MAJOR) + "." + QString::number(MINOR);
#ifdef BETA
  greeting += ".b" + QString::number(BETA);
#endif
  qmc2MainWindow->log(LOG_QMC2, greeting);

  qmc2MainWindow->log(LOG_QMC2, QObject::tr("processing global MAME configuration"));
  QHBoxLayout *layout = new QHBoxLayout;
  qmc2GlobalMAMEOptions = new MameOptions("MAME/Configuration/Global", qmc2Options->tabGlobalConfiguration);
  qmc2GlobalMAMEOptions->load();
  layout->addWidget(qmc2GlobalMAMEOptions);
  qmc2Options->tabGlobalConfiguration->setLayout(layout);
  qmc2GlobalMAMEOptions->show();

  return app.exec();
}
